package client.model;

import java.awt.Color;
import java.awt.Font;
import java.awt.event.KeyEvent;
import java.io.File;
import java.util.Random;

import org.lwjgl.LWJGLException;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.Display;

import client.Engine;
import client.Settings;
import client.twod.GLFont;
import client.twod.TwoD;
import client.util.Utils;

import com.threed.jpct.Camera;
import com.threed.jpct.Config;
import com.threed.jpct.FrameBuffer;
import com.threed.jpct.GLSLShader;
import com.threed.jpct.Loader;
import com.threed.jpct.Object3D;
import com.threed.jpct.SimpleVector;
import com.threed.jpct.Texture;
import com.threed.jpct.TextureInfo;
import com.threed.jpct.TextureManager;
import com.threed.jpct.World;
import com.threed.jpct.threading.WorkLoad;
import com.threed.jpct.threading.Worker;
import com.threed.jpct.util.KeyMapper;
import com.threed.jpct.util.KeyState;

public class Client {

	private final String fileStorage;
	private final World world;
	private final Camera camera;
	private final KeyMapper keyMapper;
	private KeyState keyState;
	private final TwoD twod;
	private final WorldMap terrain;
	private final LocalPlayer localPlayer;
	private final Object3D maleModel;
	private final Object3D femaleModel;
	private final TextureInfo maleBaseTexture;
	private final TextureInfo femaleBaseTexture;
	private final Runtime runtime = Runtime.getRuntime();
	private GLSLShader shader;
	private PropManager propManager;
	private NpcManager npcManager;
	private final GroundItemManager groundItemManager;

	/**
	 * When called, visibility list updates will be forced.
	 * Perfect for entering a new region, such as when logging in or teleporting.
	 */
	private boolean forceVisibilityUpdate;

	private final GLFont NANO_FONT = new GLFont(new Font("Consoles", Font.PLAIN, 10));
	private final GLFont GAME_FONT = new GLFont(new Font("Suego UI", Font.BOLD, 14));
	private final GLFont INTERFACE_FONT = new GLFont(new Font("Suego UI", Font.PLAIN, 18));

	@SuppressWarnings("unused")
	private final Color GUI_BACKGROUND_COLOR = new Color(0x484a40);
	@SuppressWarnings("unused")
	private final Color GUI_INTERFACE_OVERLAY_COLOR = new Color(0xcdcac2);
	private final Color GUI_TEXT_COLOR = new Color(0xfef167);

	private boolean showCursor = false;

	// Login menu state
	private boolean loginMenuOpened = false;

	// Toggle-able ui components
	private boolean inputConsole = false;
	private boolean playerHud = false;
	private boolean potionHud = false;

	// Welcome interface
	private boolean welcomeInterface ;
	private String[] welcomeMessage = { // 12 lines of text. if unused, leave empty.. "",
			"Welcome to " + Config.glWindowName + " (Alpha / Tech Demo)",
			"",
			"You will need to understand some basic controls:",
			"Hold [WASD] to move around the terrain",
			"Use your mouse to left click an icon at the bottom left.",
			"Press [F] to toggle point of view",
			"Press [CTRL] to toggle 'godspeed' cheat",
			"Press [1-4] to teleport to different areas of the map",
			"",
			"To exit the game at any time, press [ESCAPE]",
			"",
			"Press [ENTER] to continue."
	};

	private boolean dialogueInterface;
	private String[] dialogueMessage = { // 12 lines of text. if unused, leave empty.. "",
			"",
			"",
			"",
			"Empty Dialogue Interface",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
	};
	

	private boolean inventoryInterface;
	private boolean skillTreeInterface;
	private boolean objectiveInterface;
	private boolean logoutRequest;
	private long lastAppearanceUpdate = 0L;
	private int goldPieces = 9001;
	private String currentLocation = "Unknown";

	public Client(final String fileStorage) {
		this.fileStorage = fileStorage;

		this.keyMapper = new KeyMapper(); // Initialize keyboard
		// Initialize mouse
		try {
			Mouse.create();
			Mouse.setGrabbed(false);
		} catch (LWJGLException e) {
			e.printStackTrace();
		}

		// Configure jPCT
		Config.farPlane = 750;
		Config.maxPolysVisible = 36500;

		/**
		 * Initialize 2D graphics.
		 */
		this.twod = new TwoD(fileStorage);

		/**
		 * Initialize the world map.
		 */
		this.terrain = new WorldMap(fileStorage, world = new World());

		/**
		 * Load the head icons.
		 */
		for (int i = 0; i < 3; i++) {
			TextureManager.getInstance().addTexture("head_icon" + i, new Texture(fileStorage + "game" + File.separator + "head_icon" + i + ".png"));
		}
		TextureManager.getInstance().addTexture("white", new Texture(1, 1, Color.WHITE));
		TextureManager.getInstance().addTexture("splat-baked", new Texture(fileStorage + "terrain" + File.separator + "splat-baked.png"));

		// Fire particle related.
		Config.polygonBufferSize = 10;
		TextureManager.getInstance().addTexture("smoke", new Texture(fileStorage + "game" + File.separator + "smoke.jpg"));
		TextureManager.getInstance().addTexture("fire", new Texture(fileStorage + "game" + File.separator + "flame.jpg"));

		/**
		 * Build the male / female models.
		 */
		this.maleModel = Loader.loadMD2(fileStorage + "player" + File.separator + "male.md2", .10f);
		TextureManager.getInstance().addTexture("male", new Texture(fileStorage + "player" + File.separator + "male.jpg"));
		TextureManager.getInstance().addTexture("male-mask", new Texture(fileStorage + "player" + File.separator + "male.png"));
		this.maleBaseTexture = new TextureInfo(TextureManager.getInstance().getTextureID("male"));
		maleBaseTexture.add(TextureManager.getInstance().getTextureID("male-mask"), TextureInfo.MODE_ADD);
		maleModel.setTexture(maleBaseTexture);
		//maleModel.setTexture("male");
		maleModel.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		maleModel.rotateMesh();
		maleModel.compile(true); // Compile dynamic object

		this.femaleModel = Loader.loadMD2(fileStorage + "player" + File.separator + "female.md2", .091f);
		TextureManager.getInstance().addTexture("female", new Texture(fileStorage + "player" + File.separator + "female.jpg"));
		TextureManager.getInstance().addTexture("female-mask", new Texture(fileStorage + "player" + File.separator + "female.png"));
		this.femaleBaseTexture = new TextureInfo(TextureManager.getInstance().getTextureID("male"));
		femaleBaseTexture.add(TextureManager.getInstance().getTextureID("female-mask"), TextureInfo.MODE_ADD);
		femaleModel.setTexture(femaleBaseTexture);
		//femaleModel.setTexture("female");
		femaleModel.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		femaleModel.rotateMesh();
		femaleModel.compile(true); // Compile dynamic object

		/**
		 * Initialize the local player
		 */
		this.localPlayer = new LocalPlayer(fileStorage, this, world, this.camera = world.getCamera(), terrain);
		Color hair = Color.RED;
		Color bottom = Color.BLUE;
		Color top = Color.GREEN;
		float[][] randomColors = new float[][] {
			{ hair.getRed(), hair.getGreen(), hair.getBlue() },
			{ bottom.getRed(), bottom.getGreen(), bottom.getBlue() },
			{ top.getRed(), top.getGreen(), top.getBlue() }
		};
		localPlayer.updateAppearance(this, false, randomColors, world);

		//World.setDefaultThread(Thread.currentThread());

		/**
		 * Load prop models and spawns
		 */
		Worker worker = new Worker(1);
		worker.add(new WorkLoad() {
			public void doWork() {
				propManager = new PropManager(fileStorage, world, terrain);
			}
			public void done() { }
			public void error(Exception arg0) { }
		});
		worker.waitForAll();
		worker.dispose();
		worker = null;

		/**
		 * Load npc models and spawns
		 */
		worker = new Worker(1);
		worker.add(new WorkLoad() {
			public void doWork() {
				npcManager = new NpcManager(fileStorage, world, terrain);
			}
			public void done() { }
			public void error(Exception arg0) { }
		});
		worker.waitForAll();
		worker.dispose();

		this.groundItemManager = new GroundItemManager(fileStorage, world, terrain);
		
		/**
		 * TODO Load or create the player's profile.
		 */
		forceVisibilityUpdate = true;
		localPlayer.teleportTo(WorldMap.HOUSE_LOCATION);
	}

	/**
	 * 3D rendering!!
	 */
	public void render3D(final FrameBuffer frameBuffer) {
		if (loginMenuOpened) {
			frameBuffer.clear(Color.BLACK);
			return;
		}
		frameBuffer.clear(terrain.getClearColor());
		terrain.renderSkyBox(frameBuffer, twod);
		world.renderScene(frameBuffer);
		world.draw(frameBuffer);
		//world.drawWireframe(frameBuffer, Color.BLACK);
		terrain.renderLensFlare(frameBuffer);
	}

	/**
	 * 2D rendering!!
	 */
	public void render2D(final FrameBuffer frameBuffer, final long lastFrameCount, final int lastUpdateCount) {
		if (loginMenuOpened) {
			twod.drawImage(frameBuffer, "login", frameBuffer.getWidth() / 2 - 400, frameBuffer.getHeight() / 2 - 300, true);
			GAME_FONT.drawString(frameBuffer, "username_here", Display.getWidth() / 2 - 95, Display.getHeight() / 2 + 13, Color.YELLOW);
			GAME_FONT.drawString(frameBuffer, "password_here", Display.getWidth() / 2 - 95, Display.getHeight() / 2 + 74, Color.YELLOW);
		}
		NANO_FONT.drawShadowedString(frameBuffer, "Memory Usage: " + ((runtime.totalMemory() - runtime.freeMemory()) >> 20) + "M", 5, 50, GUI_TEXT_COLOR);
		NANO_FONT.drawShadowedString(frameBuffer, "Frames Per Second: " + lastFrameCount, 5, 62, GUI_TEXT_COLOR);
		NANO_FONT.drawShadowedString(frameBuffer, "Mouse x:" + Engine.getMouseX() + " y:" + Engine.getMouseY(), 5, 74, GUI_TEXT_COLOR);
		if (loginMenuOpened) {
			return;
		}
		NANO_FONT.drawShadowedString(frameBuffer, "Camera x:" + camera.getPosition().x + " y: " + camera.getPosition().y + " z:" + camera.getPosition().z, 5, 86, GUI_TEXT_COLOR);
		NANO_FONT.drawShadowedString(frameBuffer, "Player x:" + localPlayer.getX() + " y: " + localPlayer.getY() + " z:" + localPlayer.getZ(), 5, 98, GUI_TEXT_COLOR);
		NANO_FONT.drawShadowedString(frameBuffer, "Props In View: " + propManager.getVisibleEntityCount(), 5, 110, GUI_TEXT_COLOR);
		NANO_FONT.drawShadowedString(frameBuffer, "Npcs In View: " + npcManager.getVisibleEntityCount(), 5, 122, GUI_TEXT_COLOR);

		/**
		 * Render the MMO/RPG game frame!
		 */
		if (inputConsole) {
			twod.drawImage(frameBuffer, "ui_console1", 2, Display.getHeight() - (158 + 42 - 8), !Settings.PERFORMANCE_MODE);
		}
		twod.drawImage(frameBuffer, "ui_game", 0, Display.getHeight() - 42, true);
		GAME_FONT.drawShadowedString(frameBuffer, goldPieces + "", 248, Display.getHeight() - 24, GUI_TEXT_COLOR);
		GAME_FONT.drawShadowedString(frameBuffer, "Console", 244, Display.getHeight() - 5, Color.WHITE);
		GAME_FONT.drawShadowedString(frameBuffer, "Dialogue", 334, Display.getHeight() - 5, Color.WHITE);

		frameBuffer.blit(TextureManager.getInstance().getTexture("splat-baked"), 0, 0, Display.getWidth() - 136 - 10, 10, 512, 512, 136, 136, -1, true);
		twod.drawImage(frameBuffer, "ui_map", Display.getWidth() - 164, -8, true);

		/**
		 * Render the current interface!
		 */
		if (potionHud) {
			for (int i = 0; i < 4; i++) {
				twod.drawImage(frameBuffer, "potion"+i, 2, Display.getHeight() / 2 - 72 + (i * 27), true);
			}
		}
		if (playerHud) {
			for (int j = 0; j < 3; j++) {
				//twod.drawImage(frameBuffer, "overlay"+j, 2, Display.getHeight() / 2 + (j * 25) - 100, true);
				twod.drawImage(frameBuffer, "overlay"+j, Display.getWidth() - 247, Display.getHeight() / 2 - (j * 28), true);
			}
		}
		if (inventoryInterface) {
			twod.drawImage(frameBuffer, "iface2", /*Display.getWidth()*/800 / 2 - 266, /*Display.getHeight()*/600 / 2 - 175, true);
		}
		if (skillTreeInterface) {
			twod.drawImage(frameBuffer, "iface4", /*Display.getWidth()*/800 / 2 - 266, /*Display.getHeight()*/600 / 2 - 175, true);
		}
		if (objectiveInterface) {
			twod.drawImage(frameBuffer, "iface3", /*Display.getWidth()*/800 / 2 - 266, /*Display.getHeight()*/600 / 2 - 175, true);
		}
		if (logoutRequest) {
			// TODO : text popup for "click the logout button again to confirm"
		}
		if (welcomeInterface || dialogueInterface) {
			twod.drawImage(frameBuffer, "iface3", Display.getWidth() / 2 - 266, Display.getHeight() / 2 - 175, false);
			if (welcomeInterface) {
				for (int i = 0; i < welcomeMessage.length; i++) {
					INTERFACE_FONT.drawShadowedString(frameBuffer, welcomeMessage[i], Display.getWidth() / 2 - 200, Display.getHeight() / 2 - 110 + (i * 20), GUI_TEXT_COLOR);
				}
			} else if (dialogueInterface) {
				for (int i = 0; i < dialogueMessage.length; i++) {
					INTERFACE_FONT.drawShadowedString(frameBuffer, dialogueMessage[i], Display.getWidth() / 2 - 200, Display.getHeight() / 2 - 110 + (i * 20), GUI_TEXT_COLOR);
				}
			}
		}

		currentLocation = "Open Terrain";

		if (localPlayer.getX() > 9225 && localPlayer.getX() < 9800 && localPlayer.getZ() > 10077 && localPlayer.getZ() < 10200) {
			currentLocation = "Lumber Ridge"; // trail to castle
		}
		if (localPlayer.getX() > 8926 && localPlayer.getX() < 9165 && localPlayer.getZ() > 9996 && localPlayer.getZ() < 10432) {
			currentLocation = "Lumber Ridge Castle";
		}
		if (localPlayer.getX() > 8197 && localPlayer.getX() < 11803 && localPlayer.getZ() > 8198 && localPlayer.getZ() < 9039) {
			twod.drawImage(frameBuffer, "pvp", Display.getWidth() / 2 - 64, 0, true);
			currentLocation = "Uncharted Region (PvP)";
		}
		if (localPlayer.getX() > 9947 && localPlayer.getX() < 10070 && localPlayer.getZ() > 9933 && localPlayer.getZ() < 10050) {
			currentLocation = "Center of the Universe";
		}
		if (localPlayer.getX() > 9812 && localPlayer.getX() < 10137 && localPlayer.getZ() > 10386 && localPlayer.getZ() < 10809) {
			// this is just for the 4 npc houses
			currentLocation = "Lumber Ridge Village";
		}
		if (localPlayer.getX() > 9436 && localPlayer.getX() < 9593 && localPlayer.getZ() > 9542 && localPlayer.getZ() < 9745) {
			currentLocation = "Home";
		}
		if (localPlayer.getX() > 9637 && localPlayer.getX() < 9732 && localPlayer.getZ() > 10316 && localPlayer.getZ() < 10430) {
			currentLocation = "General Shop";
		}
		if (localPlayer.getX() > 10776 && localPlayer.getX() < 11793 && localPlayer.getZ() > 11344 && localPlayer.getZ() < 11791) {
			currentLocation = "Goblin Hideout";
		}
		if (Utils.isInRange(275, localPlayer.getFocusPoint().getTransformedCenter(), WorldMap.BURNING_ISLAND)) {
			currentLocation = "Burning Island";
		}
		if (Utils.isInRange(126, localPlayer.getFocusPoint().getTransformedCenter(), WorldMap.CAVE_ENTRANCE)) {
			currentLocation = "Cave Entrance";
		}
		// TODO : add a lot more areas! this is an original feature. and fits nicely with the
		// open terrain setup.
		if (localPlayer.getX() > 0 && localPlayer.getX() < 0 && localPlayer.getZ() > 0 && localPlayer.getZ() < 0) {
		}
		if (localPlayer.getX() > 0 && localPlayer.getX() < 0 && localPlayer.getZ() > 0 && localPlayer.getZ() < 0) {
		}
		if (localPlayer.getX() > 0 && localPlayer.getX() < 0 && localPlayer.getZ() > 0 && localPlayer.getZ() < 0) {
		}
		if (localPlayer.getX() > 0 && localPlayer.getX() < 0 && localPlayer.getZ() > 0 && localPlayer.getZ() < 0) {
		}
		if (localPlayer.getX() > 0 && localPlayer.getX() < 0 && localPlayer.getZ() > 0 && localPlayer.getZ() < 0) {
		}
		if (localPlayer.getX() > 0 && localPlayer.getX() < 0 && localPlayer.getZ() > 0 && localPlayer.getZ() < 0) {
		}

		GAME_FONT.drawShadowedString(frameBuffer, "Location: " + currentLocation, 5, 15, GUI_TEXT_COLOR);

		int mouseX = Mouse.getX();
		int mouseY = (frameBuffer.getOutputHeight() - Mouse.getY());
		if (showCursor) {
			twod.drawImage(frameBuffer, "cursor", mouseX - 16, mouseY - 16, true);
		}
		/**
		 * Poll mouse input for the game frame.
		 */
		while (Mouse.next()) {
			if (Mouse.getEventButton() > -1) {
				if (Mouse.getEventButtonState()) {
					//System.out.println("PRESSED MOUSE BUTTON: " + Mouse.getEventButton());
				} else {
					//System.out.println("RELEASED MOUSE BUTTON: " + Mouse.getEventButton());
					/**
					 * The ui buttons
					 */
					for (int i = 0; i < 5; i++) {
						if (
								mouseX > (i * 45) &&
								mouseX < 42 + (i * 45) &&
								mouseY > Display.getHeight() - 40 &&
								mouseY < Display.getHeight()
								) {
							System.out.println("ui button #"+i);
							switch (i) {
							case 0: // inventory
								inventoryInterface = !inventoryInterface;
								break;
							case 1: // character
								if (lastAppearanceUpdate <= System.currentTimeMillis() - 1500) {
									lastAppearanceUpdate = System.currentTimeMillis();
									float[][] randomColors = new float[][] { { new Random().nextFloat(), new Random().nextFloat(), new Random().nextFloat() }, { new Random().nextFloat(), new Random().nextFloat(), new Random().nextFloat() }, { new Random().nextFloat(), new Random().nextFloat(), new Random().nextFloat() } };
									localPlayer.updateAppearance(this, false, randomColors, world);
								}
								break;
							case 2: // skill tree
								skillTreeInterface = !skillTreeInterface;
								break;
							case 3: // adventure log
								objectiveInterface = !objectiveInterface;
								break;
							case 4: // logout
								logoutRequest = true;
								break;
							}
						} // end ui buttons
						/**
						 * The console buttons
						 */
						for (int j = 0; j < 2; j++) {
							if (
									mouseX > 227 + (j * 90) && 
									mouseX < 227 + ((j + 1) * 90) && 
									mouseY > Display.getHeight() - 20 && 
									mouseY < Display.getHeight()
									) {
								if (j == 0) {
									inputConsole = !inputConsole;
								}
								if (j == 1) {
									//dialogueconsole = !dialogueConsole;
								}
							}
						} // end console buttons
						/**
						 * The inventory interface buttons
						 */
						if (inventoryInterface) {
							// show inventory
							// 532 350
							for (int j = 0; j < 4; j++) { // row 1
								// 207, 149
								// 287, 224
								// gap between each button = 15px
								// button size = 80px
								int m = (j * 82) + (j * 17);
								if (mouseX > 207 + m && mouseX < 287 + m && mouseY > 149 && mouseY < 224) {
									// valid clicking coordinates of each item slot
									System.out.println("inventory debug: row 1, slot " + (j + 1));
								}
							}
							for (int j = 0; j < 4; j++) { // row 2
								// 207, 249
								// 287, 324
								int m = (j * 82) + (j * 17);
								if (mouseX > 207 + m && mouseX < 287 + m && mouseY > 249 && mouseY < 324) {
									// valid clicking coordinates of each item slot
									System.out.println("inventory debug: row 2, slot " + (j + 1));
								}
							}
							for (int j = 0; j < 4; j++) { // row 3
								// 207, 349
								// 287, 424
								int m = (j * 82) + (j * 17);
								if (mouseX > 207 + m && mouseX < 287 + m && mouseY > 349 && mouseY < 424) {
									// valid clicking coordinates of each item slot
									System.out.println("inventory debug: row 3, slot " + (j + 1));
								}
							}
							if (mouseX > 343 && mouseX < 454 && mouseY > 442 && mouseY < 462) {
								// close the interface
								inventoryInterface = false;
							}
						} // end inventory clicking

					}
				}
			}
		}

	}

	/**
	 * Update the game!!
	 */
	public void update(final float delta, final FrameBuffer frameBuffer) {
		keyState = null;
		/**
		 * Poll keyboard input first.
		 */
		if ((keyState = keyMapper.poll()) != KeyState.NONE) {
			/**
			 * Terminate the application!
			 */
			switch (keyState.getKeyCode()) {
			case KeyEvent.VK_ESCAPE:
				Engine.terminate();
				break;
			}
			if (loginMenuOpened) {
				/**
				 * Handle login input
				 */
				switch (keyState.getKeyCode()) {
				case KeyEvent.VK_ENTER:
					loginMenuOpened = false;
					break;
				}
			} else {
				/**
				 * Handle game input
				 */
				if (keyState.getKeyCode() == KeyEvent.VK_ENTER && keyState.getState()) {
					if (welcomeInterface) {
						welcomeInterface = false;
					}
				}
				if (isInterfaceOpened()) {
					return;
				}
				switch (keyState.getKeyCode()) {
				case KeyEvent.VK_0:
					if (keyState.getState()) {
						Engine.screenShot();
					}
					break;
				case KeyEvent.VK_1:
					localPlayer.teleportTo(WorldMap.BURNING_ISLAND);
					break;
				case KeyEvent.VK_2:
					localPlayer.teleportTo(WorldMap.BURNING_DUNGEON);
					break;
				case KeyEvent.VK_3: // goblin village
					localPlayer.teleportTo(new SimpleVector(11532.492, 90.9, 11606.823));
					break;
				case KeyEvent.VK_4:
					localPlayer.teleportTo(WorldMap.CENTER_OF_THE_UNIVERSE);
					break;
				case KeyEvent.VK_TAB:
					if (keyState.getState()) {
						inputConsole = !inputConsole;
					}
					break;
				case KeyEvent.VK_I:
					if (keyState.getState()) {
						groundItemManager.spawnRandomItem(localPlayer.getFocusPoint().getTransformedCenter());
					}
					break;
				case KeyEvent.VK_ENTER:
					if (keyState.getState()) {
						//int hardcode_id = 18;// = new Random().nextInt(13) == 6 ? 19 : 20;
						//System.out.println(hardcode_id + ", " + localPlayer.getX() + ", " + localPlayer.getY() + ", " + localPlayer.getZ() + " // ");

						GroundItem pickUp = groundItemManager.pickUp(localPlayer.getFocusPoint());
						if (pickUp != null) {
							System.out.println("picked up item (type #"+pickUp.getType()+", uid #"+pickUp.getUid()+")");
							pickUp.expire(this);
						}
						
						/**
						 * Check if the interaction is with an NPC
						 */
						if (npcManager.getInteractionEntity(localPlayer.getFocusPoint()) != null) {
							//Npc npc = npcManager.getInteractionEntity(localPlayer.getFocusPoint());
							//npc.setLeader(localPlayer.getFocusPoint());
							dialogueInterface = true;
							dialogueMessage[0] = "The NPC has nothing interesting to say..";
						}
						/**
						 * Check if the interaction is with a dynamic prop
						 */
						if (propManager.getInteractionEntity(localPlayer.getFocusPoint()) != null) {
							Prop prop = propManager.getInteractionEntity(localPlayer.getFocusPoint());
							if (Utils.isInRange(30, localPlayer.getFocusPoint().getTransformedCenter(), prop.getTransformedCenter())) {
								switch (prop.getType()) {
								case 0:
								case 1:
								case 12:
								case 13:
								case 19:
								case 20: // trees
									System.out.println("It's a tree.");
									break;
								case 15:
									System.out.println("It's a fire pit.");
									break;
								case 23:
								case 24:
									System.out.println("It's a grave stone.");
									break;
								case 21:
									System.out.println("It's a bed.");
									break;
								case 2:
									System.out.println("Dragon cave entrance.");
									break;
								case 3:
									System.out.println("It's an item chest.");
									break;
								case 4:
									System.out.println("It's a portal.");
									break;
								case 8:
									System.out.println("The general shop.");
									break;
								}
							}
						}
					}
					break;
				}
			}
		}

		/**
		 * Update the list of entities in view 10x / minute
		 */
		propManager.updateVisibilityList(camera, forceVisibilityUpdate);
		propManager.cycle(delta);
		npcManager.updateVisibilityList(camera, forceVisibilityUpdate);
		npcManager.cycle(delta);
		groundItemManager.updateVisibilityList(camera, forceVisibilityUpdate);
		groundItemManager.cycle(delta);
		forceVisibilityUpdate = false;

		/**
		 * Update the local player
		 */
		localPlayer.update(delta, keyState);
	}

	/**
	 * Returns a fe/male object.
	 */
	public Object3D generateCharacterModel(final boolean female, final float[][] colorModifiers) {
		final Object3D object;
		if (female) {
			object = new Object3D(femaleModel, true);
			object.shareCompiledData(femaleModel);
		} else {
			object = new Object3D(maleModel, true);
			object.shareCompiledData(maleModel);
		}
		this.shader = new GLSLShader(Loader.loadTextFile(fileStorage + "player" + File.separator + "player.vert"), Loader.loadTextFile(fileStorage + "player" + File.separator + "player.frag")) {
			@Override
			public void setCurrentObject3D(Object3D obj) {
				setUniform("colorMul0", colorModifiers[0]);
				setUniform("colorMul1", colorModifiers[1]);
				setUniform("colorMul2", colorModifiers[2]);
				setStaticUniform("map0", 0);
				setStaticUniform("map1", 1);
			}
		};
		object.setRenderHook(shader);
		return object;
	}

	private boolean isInterfaceOpened() {
		return welcomeInterface || inventoryInterface || objectiveInterface || skillTreeInterface;
	}

}