package client.model;

import java.awt.event.KeyEvent;

import org.lwjgl.opengl.Display;

import client.util.Utils;

import com.threed.jpct.Camera;
import com.threed.jpct.Object3D;
import com.threed.jpct.SimpleVector;
import com.threed.jpct.World;
import com.threed.jpct.util.KeyState;

public class LocalPlayer {

	@SuppressWarnings("unused")
	private final String fileStorage;
	private final Camera camera;

	private final SimpleVector ellipsoid = new SimpleVector(4.2f, 9f, 4.2f); // The "size" of the player, for collision events.
	private final float THIRD_PERSON_CAMERA_ZOOM = 90;
	private final float THIRD_PERSON_CAMERA_HEIGHT = 45;
	private final float FIRST_PERSON_CAMERA_ZOOM = -2;
	private final float FIRST_PERSON_CAMERA_HEIGHT = ellipsoid.y; // "eye level"
	private final float DAMPING = 0.1f;
	private final float MINIMUM_SPEED = 1.2f;
	private final float MAXIMUM_SPEED = MINIMUM_SPEED + .6f;
	private final float GOD_SPEED = MINIMUM_SPEED * 5;
	private final float ROTATION_SPEED = 1.6f;
	private boolean firstPerson = false;

	// potential multi player variables
	private float currentRotation = (float) Math.PI * 1.5f; // Properly rotate the MD2 model
	private float currentSpeed = 0f;

	private boolean w = false;
	private boolean s = false;
	private boolean a = false;
	private boolean d = false;
	private boolean moved = false;
	private boolean godspeed = false;
	private SimpleVector moveRes = new SimpleVector(0, 0, 0);

	/**
	 * On the fly player model generation:
	 */
	private Object3D focusPoint; // Represents the local player's chosen model!!
	@SuppressWarnings("unused")
	private float colorModifiers[][]; // Stores this player's appearance data (colors: skin tone, shirt, boots, pants, hair, etc)

	// Animation Related
	private int currentAnimation = 1; // idle
	private long lastAnimationUpdate = 0L;
	private float animationIndex;

	private final WorldMap terrain;

	public LocalPlayer(final String fileStorage, final Client client, final World world, final Camera camera, final WorldMap terrain) {
		this.fileStorage = fileStorage;
		this.camera = camera;
		this.terrain = terrain;
		if (Display.getWidth() > 1024) {
			camera.setFOV(5f);
		} else if (Display.getWidth() > 800) {
			camera.setFOV(3f);
		} else {
			camera.setFOV(1.85f); // default
		}
		camera.moveCamera(Camera.CAMERA_MOVEOUT, firstPerson ? FIRST_PERSON_CAMERA_ZOOM : THIRD_PERSON_CAMERA_ZOOM);
		camera.moveCamera(Camera.CAMERA_MOVEUP, firstPerson ? FIRST_PERSON_CAMERA_HEIGHT : THIRD_PERSON_CAMERA_HEIGHT);
	}

	public void update(final float delta, final KeyState keyState) {
		moved = false;
		if (keyState.getKeyCode() == KeyEvent.VK_W || keyState.getKeyCode() ==  KeyEvent.VK_UP) {
			w = keyState.getState();
		}
		if (keyState.getKeyCode() == KeyEvent.VK_S || keyState.getKeyCode() ==  KeyEvent.VK_DOWN) {
			s = keyState.getState();
		}
		if (keyState.getKeyCode() == KeyEvent.VK_A || keyState.getKeyCode() ==  KeyEvent.VK_LEFT) {
			a = keyState.getState();
		}
		if (keyState.getKeyCode() == KeyEvent.VK_D  || keyState.getKeyCode() ==  KeyEvent.VK_RIGHT) {
			d = keyState.getState();
		}
		if (keyState.getKeyCode() == KeyEvent.VK_CONTROL && keyState.getState()) {
			godspeed = !godspeed; // toggle god speed
		}
		if (keyState.getKeyCode() == KeyEvent.VK_F && keyState.getState()) {
			firstPerson = !firstPerson; // why not just give the player this option? :)
		}
		// move the focusPoint
		if (w || s || a || d) {
			moved = true;
		}
		if (w) {
			if (currentSpeed < MAXIMUM_SPEED) {
				currentSpeed += 0.002f;
				if (currentSpeed > MAXIMUM_SPEED) {
					currentSpeed = MAXIMUM_SPEED;
				}
				if (godspeed) {
					currentSpeed = GOD_SPEED;
				}
			}
			SimpleVector t = focusPoint.getZAxis();
			t.scalarMul(currentSpeed);
			moveRes.add(t);
		} else {
			currentSpeed = MINIMUM_SPEED;
		}
		if (s) {
			SimpleVector t = focusPoint.getZAxis();
			t.scalarMul(-MINIMUM_SPEED);
			moveRes.add(t);
		}
		if (a) {
			currentRotation += (float) Math.toRadians(-ROTATION_SPEED);
			focusPoint.rotateY((float) Math.toRadians(-ROTATION_SPEED));
		}
		if (d) {
			currentRotation += (float) Math.toRadians(ROTATION_SPEED);
			focusPoint.rotateY((float) Math.toRadians(ROTATION_SPEED));
		}
		// avoid high speeds
		if (moveRes.length() > MAXIMUM_SPEED) {
			moveRes.makeEqualLength(new SimpleVector(0, 0, godspeed ? GOD_SPEED : MAXIMUM_SPEED));
		}

		boolean burning = false;
		if (focusPoint.getTransformedCenter().y > 100 && WorldMap.isInPvP(focusPoint)) {
			burning = true;
		}
		boolean swimming = false;
		if (focusPoint.getTransformedCenter().y > 114 && !WorldMap.isInPvP(focusPoint)) {
			swimming = true;
		}
		
		if (moved && !swimming) {
			focusPoint.translate(0, -0.025f, 0); // prevent sinking through terrain
		}

		moveRes = focusPoint.checkForCollisionEllipsoid(moveRes, ellipsoid, 8);
		focusPoint.translate(moveRes);

		// go down (gravity)
		if (!swimming) {
			SimpleVector t = new SimpleVector(0, 1, 0);
			t = focusPoint.checkForCollisionEllipsoid(t, ellipsoid, 1);
			focusPoint.translate(t);
		}

		// damping
		if (moveRes.length() > DAMPING) {
			moveRes.makeEqualLength(new SimpleVector(0, 0, DAMPING));
		} else {
			moveRes = new SimpleVector(0, 0, 0);
		}

		// update camera position
		camera.setPositionToCenter(focusPoint);
		camera.align(focusPoint); // align the camera with the player
		camera.rotateCameraX((float) Math.toRadians(15));
		if (!firstPerson) {
			camera.moveCamera(Camera.CAMERA_MOVEOUT, THIRD_PERSON_CAMERA_ZOOM);
		} else {
			camera.moveCamera(Camera.CAMERA_MOVEUP, FIRST_PERSON_CAMERA_HEIGHT); // force specific camera height
			camera.moveCamera(Camera.CAMERA_MOVEIN, FIRST_PERSON_CAMERA_ZOOM); // move orb to the front of the player's face
		}

		/**
		 * Animate the model:
		 */
		if (lastAnimationUpdate <= System.currentTimeMillis() - 10) {
			lastAnimationUpdate = System.currentTimeMillis();
			/**
			1: Stand
			2: Walk
			3: Throw
			4: Been hit
			5: Death
			6: Spell
			7: Pick from floor
			8: Eat something
			9: Floating
			 */

			if (burning) {
				currentAnimation = 4;
			} else if (focusPoint.getTransformedCenter().y > 112) {
				currentAnimation = 9;
			} else {
				if (moved) {
					currentAnimation = 2; // walk
				} else {
					currentAnimation = 1; // idle
				}
			}
			// Perform the jPCT animation:
			animationIndex += 0.02f;
			if (animationIndex > 1) {
				animationIndex -= 1;
			}
			focusPoint.animate(animationIndex, currentAnimation);
		}
	}

	public void updateAppearance(final Client client, final boolean female, final float[][] colorModifiers, final World world) {
		this.colorModifiers = colorModifiers;
		SimpleVector location = new SimpleVector();
		if (focusPoint != null) {
			location = focusPoint.getTransformedCenter();
			world.removeObject(focusPoint);
			focusPoint = null;
		}
		this.focusPoint = client.generateCharacterModel(female, colorModifiers); // Update gender
		world.addObject(focusPoint);
		focusPoint.setCollisionMode(Object3D.COLLISION_CHECK_SELF);
		focusPoint.compile(true);

		// Stop the player from bouncing around after an apperance update..
		focusPoint.clearTranslation(); // Clear current location
		focusPoint.translate(location); // Teleport

		focusPoint.clearRotation();
		focusPoint.rotateY(currentRotation);
	}

	/**
	 * Teleports the local player anywhere, and executes a map visibility update.
	 */
	public void teleportTo(final SimpleVector destination) {
		focusPoint.clearTranslation(); // Clear current location
		focusPoint.translate(destination); // Teleport
		focusPoint.translate(0, -ellipsoid.y, 0); // Go up, because sometimes we get stuck inside the terrain..??
		Utils.dropEntityToGround(focusPoint, terrain.getMapChunk(focusPoint)); // Clamp to ground
		focusPoint.translate(0, -ellipsoid.y, 0); // Go up, because sometimes we get stuck inside the terrain..??
		terrain.updateMapChunkVisibility(focusPoint); // Update map chunk visibility
	}

	/**
	 * Returns X coordinate in 3D world space.
	 */
	public float getX() {
		return focusPoint.getTransformedCenter().x;
	}

	/**
	 * Returns Y (height) coordinate in 3D world space.
	 */
	public float getY() {
		return focusPoint.getTransformedCenter().y;
	}

	/**
	 * Returns Z coordinate in 3D world space.
	 */
	public float getZ() {
		return focusPoint.getTransformedCenter().z;
	}

	/**
	 * Returns the focusPoint instance.
	 * @note The instance represents the player's model!
	 */
	public Object3D getFocusPoint() {
		return focusPoint;
	}

	public float getCurrentRotation() {
		return currentRotation;
	}

}
